import pgen, sys, string, os


ExampleGrammar = None


def compile(file):
    global ErrorCount, ExampleGrammar
    ErrorCount = 0

    if not ExampleGrammar:
	ExampleGrammar = _Init()

    try:
	tree = ExampleGrammar.parse_file(file)
    except SyntaxError, msg:
	ReportError(msg[0], msg[1])
	return 1, []
    except IOError, msg:
	ReportError(msg[1], (file, 0))
	return 1, []

    return tree


#
# Anything from the tokenizer
#
def terminal(g, tok, where, child):
    return (tok, child)


#
#% rc_file: (NEWLINE | stmt)* ENDMARKER
#
def rc_file(g, tok, where, child):
    d = {}
    for t, n, v in filter(lambda x, y=g.NT.stmt: x[0] == y, child):
        d[n] = v
    return d

    
#
#% stmt: name ':' value NEWLINE
#
def stmt(g, tok, where, child):
    return (tok, child[0][1], child[2][1])


#    
#% name: NAME ('.' NAME)*
#
def name(g, tok, where, child):
    names = filter(lambda c, t=g.T.NAME: c[0] == t, child)
    return (tok, tuple(map(lambda x: x[1], names)))
    
#    
#% list: '(' value (',' value)* [','] ')'
#
def list(g, tok, where, child):
    l = []
    for i in child[1:len(child)-1]:
	if i[0] != g.T.COMMA:
	    l.append(i[1])
    return (tok, l)


#    
#% number: [ '-' | '+' ] NUMBER
#
def number(g, tok, where, child):
    if len(child) == 2:
	return (tok, eval(child[0][1] + child[1][1]))
    else:
	return (tok, eval(child[0][1]))


#    
#% value: NAME | STRING | number | list
#
def value(g, tok, where, child):
    ch = child[0]
    t = ch[0]
    if t == g.T.NAME:
	return (tok, ch[1])
    elif t == g.NT.list:
	return (tok, ch[1])
    elif t == g.NT.number:
	return (tok, ch[1])
    else:
	return (tok, eval(ch[1]))

	
#
# Error reporting
#
ErrorCount = 0

def ReportError(msg, where):
    global ErrorCount
    ErrorCount = ErrorCount + 1
    print "%s:%d: %s" % (where[0], where[1], msg)

#    
# Module initialization
#


Rules = """
rc_file: (NEWLINE | stmt)* ENDMARKER
stmt: name ':' value NEWLINE
name: NAME ('.' NAME)*
list: '(' value (',' value)* [','] ')'
number: [ '-' | '+' ] NUMBER
value: NAME | STRING | number | list
"""

def _Init():
    parser = pgen.parser_from_string(Rules)
    if not parser:
	return None

    this_module = sys.modules[__name__]
    for i in dir(parser.NT):
	try:
	    parser.reduction[getattr(parser.NT, i)] = getattr(this_module, i)
	except AttributeError:
	    parser.reduction[getattr(parser.NT, i)] = terminal

    for i in dir(parser.T):
	try:
	    parser.reduction[getattr(parser.T, i)] = getattr(this_module, i)
	except AttributeError:
	    parser.reduction[getattr(parser.T, i)] = terminal

    return parser


if __name__  == "__main__":
    print compile('test.rc')
