/***********************************************************
 Copyright 1991, 1992, 1993, 1994 by Stichting Mathematisch Centrum,
                     Amsterdam, The Netherlands.

                         All Rights Reserved

 Permission to use, copy, modify, and distribute this software and its
 documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and that
 both that copyright notice and this permission notice appear in
 supporting documentation, and that the names of Stichting Mathematisch
 Centrum or CWI not be used in advertising or publicity pertaining to
 distribution of the software without specific, written prior
 permission.

 STICHTING MATHEMATISCH CENTRUM DISCLAIMS ALL WARRANTIES WITH REGARD TO
 THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS, IN NO EVENT SHALL STICHTING MATHEMATISCH CENTRUM BE LIABLE
 FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

/*******************************************************************
   Copyright 1993, 1994 by V.I. Corporation, Northampthon, MA, USA
    
                         All Rights Reserved
                      
 Permission to use, copy, modify, and distribute this software and its
 documentation for any purpose and without fee is hereby granted,
 provided that the above copyright notice appear in all copies and
 that both that copyright notice and this permission notice appear in
 supporting documentation, and that the name of V.I. Corporation not
 be used in advertising or publicity pertaining to distribution of the
 software without specific, written prior permission.
 
 V.I. CORPORATION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS, IN NO EVENT SHALL V.I. CORPORATION BE LIABLE FOR ANY
 SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

**********************************************************************/

/*******************************************************************
       Copyright 1997 by Donald Beaudry, Northborough, MA, USA
    
                         All Rights Reserved
                      
 Permission to use, copy, modify, and distribute this software and
 its documentation for any purpose and without fee is hereby
 granted, provided that the above copyright notice appear in all
 copies and that both that copyright notice and this permission
 notice appear in supporting documentation, and that the name of
 Donald Beaudry not be used in advertising or publicity pertaining
 to distribution of the software without specific, written prior
 permission.
 
 DONALD BEAUDRY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 NO EVENT SHALL DONALD BEAUDRY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
 USE OR PERFORMANCE OF THIS SOFTWARE.

**********************************************************************/

#include <assert.h>

#include "Python.h"
#include "compile.h"
#include "structmember.h"
#include "node.h"
#include "token.h"
#include "pythonrun.h"
#include "graminit.h"
#include "grammar.h"
#include "errcode.h"
#include "parsetok.h"
#include "pgenheaders.h"

typedef struct
{  
  grammar *grammar;
  node *tree;
  PyObject *nonterminals;
  PyObject *nonterminal_list;
  PyObject *keywords;
  PyObject *reductions;
} GrammarPart;


typedef struct 
{
  PyObject_HEAD
  GrammarPart gram;
} GrammarObject;


static void Dealloc Py_PROTO((GrammarObject *go));
static PyObject *CreateGrammar Py_PROTO((node *n));
static PyObject *ParseFile Py_PROTO((GrammarObject *go, PyObject *args));
static PyObject *ParseString Py_PROTO((GrammarObject *go, PyObject *args));
static PyObject *PrintGrammar Py_PROTO((GrammarObject *go, PyObject *args));
static PyObject *PrintNonterminals Py_PROTO((GrammarObject *go, PyObject *args));
static PyObject *GetAttr Py_PROTO((GrammarObject *go, char *name));
static PyObject *ParseGrammarFile Py_PROTO((PyObject *self, PyObject *args));
static PyObject *ParseGrammarString Py_PROTO((PyObject *self, PyObject *args));


PyTypeObject
Grammar_Type =
{
  PyObject_HEAD_INIT(&PyType_Type)
    0,				/* ob_size */
    "pgen.grammar",		/* tp_name */
    sizeof (GrammarObject),     /* tp_size */
    0,				/* tp_itemsize */
    /* methods */
    (destructor)Dealloc,        /* tp_dealloc */
    0,                          /* tp_print */
    (getattrfunc)GetAttr,       /* tp_getattr */
    0,                          /* tp_setattr */
    0,                          /* tp_compare */
    0,                          /* tp_repr */
    0,                          /* tp_as_number */
    0,				/* tp_as_sequence */
    0,				/* tp_as_mapping */
};


static PyMethodDef GlobalMethods[] =
{
  {"parser_from_file", (PyCFunction)ParseGrammarFile, METH_VARARGS},
  {"parser_from_string", (PyCFunction)ParseGrammarString, METH_VARARGS},
  {0, 0} /* Sentinel */
};


static PyMethodDef GrammarMethods[] =
{
  {"parse_file", (PyCFunction)ParseFile, METH_VARARGS},
  {"parse_string", (PyCFunction)ParseString, METH_VARARGS},
  {"print_grammar", (PyCFunction)PrintGrammar, METH_VARARGS},
  {"print_nonterminals", (PyCFunction)PrintNonterminals, METH_VARARGS},
  
  {NULL, NULL},
};

static PyObject *TerminalEnum = 0;
static PyObject *TerminalList = 0;


static PyObject *
CreateEnum(name, name_list)
     char *name;
     PyObject *name_list;
{
  PyObject *mo;
  PyObject *dict;
  int i;
  int n_names = PySequence_Length(name_list);
	
  mo = PyModule_New(name);
  dict = PyModule_GetDict(mo);
  
  for (i = 0; i < n_names; i++)
    {
      PyObject *name = PySequence_GetItem(name_list, i);
      PyDict_SetItem(dict, name, name);
      Py_DECREF(name);
    }

  {
    PyObject *n = PyString_FromString("__members__");
    PyObject *l = PySequence_Tuple(name_list);
    PyDict_SetItem(dict, n, l);
    Py_DECREF(n);
    Py_DECREF(l);
  }
  
  return (PyObject*) mo;
}


static PyObject *
CreateNonterminalList(g)
     grammar *g;
{
  PyObject *l;
  dfa *d;
  int i;

  l = PyTuple_New(g->g_ndfas);

  d = g->g_dfa;
  for (i = g->g_ndfas; --i >= 0; d++)
    {
      PyObject *name;
      int token;

      name = PyString_FromString(d->d_name);
      token = d->d_type - NT_OFFSET;
      
      PyTuple_SetItem(l, token, name);
    }

  return l;
}


static PyObject *
CreateKeywordList(g)
     grammar *g;
{
  PyObject *l;
  label *ll;
  int i;

  l = PyList_New(0);

  ll = g->g_ll.ll_label+1;
  for (i = 1; i < g->g_ll.ll_nlabels; ++ll, ++i)
    {
      if (ll->lb_str)
        {
          PyObject *name;
          name = PyString_FromString(ll->lb_str);
          PyList_Append(l, name);
        }
    }

  return l;
}


static PyObject *
CreateTerminalList(name_list, n_names)
     char **name_list;
     int n_names;
{
  PyObject *l;
  int i;

  l = PyTuple_New(n_names);
  
  for (i = 0; i < n_names; i++)
    {
      PyObject *name = PyString_FromString(name_list[i]);
      PyTuple_SetItem(l, i, name);
    }

  return l;
}



static PyObject *
Where(filename, line)
     PyObject *filename;
     int line;
{
  static PyObject *where = 0;
  static PyObject *fn = 0;
  static int ln = 0;

  if (filename == fn && line == ln)
    return where;

  Py_XDECREF(where);
  fn = filename;
  ln = line;
  where = Py_BuildValue("(Oi)", filename, line);
  return where;
}


static PyObject *
Node2Tuple(go, filename, n)
     GrammarObject *go;
     PyObject *filename;
     node *n;
{
  if (n == NULL)
    {
      Py_INCREF(Py_None);
      return Py_None;
    }

  if (ISTERMINAL(TYPE(n)))
    {
      PyObject *token;
      PyObject *res;
      PyObject *reduction;
      PyObject *reduce;
      
      token = PySequence_GetItem(TerminalList, TYPE(n));
      if (! token)
        {
          assert(PyErr_Occurred());
          return NULL;
        }

      reduce = PyDict_GetItem(go->gram.reductions, token);
      if (! reduce)
        {
          PyErr_Clear();

          res = Py_BuildValue("(OOs)",
                              token, Where(filename, n->n_lineno), STR(n));
          Py_DECREF(token);
          
          assert(res != NULL || PyErr_Occurred());
          return res;
        }

      res = Py_BuildValue("(OOOs)",
                          go, token, Where(filename, n->n_lineno), STR(n));
      Py_DECREF(token);

      reduction = PyEval_CallObject(reduce, res);
      Py_XDECREF(res);
      
      assert(reduction != NULL || PyErr_Occurred());
      return reduction;
    }
  

  if (ISNONTERMINAL(TYPE(n)))
    {
      int i;
      PyObject *v, *w;
      PyObject *children;
      PyObject *reduce;
      PyObject *reduction;
      PyObject *token;

      children = PyTuple_New(NCH(n));
      if (children == NULL)
        {
          assert(PyErr_Occurred());
          return NULL;
        }

      for (i = 0; i < NCH(n); i++)
        {
          w = Node2Tuple(go, filename, CHILD(n, i));
          if (w == NULL)
            {
              PyObject *err, *val, *tb;

              PyErr_Fetch(&err, &val, &tb);
              Py_XDECREF(children);
              PyErr_Restore(err, val, tb);
              assert(PyErr_Occurred());
              return NULL;
            }

          PyTuple_SetItem(children, i, w);
        }
      
      token = PySequence_GetItem(go->gram.nonterminal_list, TYPE(n)-NT_OFFSET);
      if (token == NULL)
        {
          Py_DECREF(children);
          assert(PyErr_Occurred());
          return NULL;
        }

      reduce = PyDict_GetItem(go->gram.reductions, token);
      if (! reduce)
        {
          PyErr_Clear();

          v = Py_BuildValue("(OOO)",
                            token, Where(filename, n->n_lineno), children);
          Py_DECREF(token);
          Py_DECREF(children);
          assert(v != NULL || PyErr_Occurred());
          return v;
        }

      v = Py_BuildValue("(OOOO)",
                        go, token, Where(filename, n->n_lineno), children);
      Py_DECREF(token);
      Py_DECREF(children);
      if (v == NULL)
        {
          assert(PyErr_Occurred());
          return NULL;
        }

      reduction = PyEval_CallObject(reduce, v);
      if (! reduction)
        fprintf(stderr, "%s:%d:%s\n",
                PyString_AsString(filename), n->n_lineno,
                PyString_AsString(token));
      Py_DECREF(v);
      
      return reduction;
    }

  PyErr_SetString(PyExc_SystemError, "unrecognized parse tree node type");
  return NULL;
}


static PyObject *
ParseError(err)
     perrdetail *err;
{
  PyObject *v, *w;
  char *msg = NULL;

  v = Py_BuildValue("(ziiz)", err->filename,  err->lineno,
                    err->offset, err->text); 
  if (err->text != NULL)
    {
      free(err->text);
      err->text = NULL;
    }
  
  switch (err->error)
    {
    case E_SYNTAX:
      msg = "invalid syntax";
      break;

    case E_TOKEN:
      msg = "invalid token";
      break;

    case E_INTR:
      PyErr_SetNone(PyExc_KeyboardInterrupt);
      return NULL;;

    case E_NOMEM:
      PyErr_NoMemory();
      return NULL;

    case E_EOF:
      msg = "unexpected EOF while parsing";
      break;

    default:
      fprintf(stderr, "error=%d\n", err->error);
      msg = "unknown parsing error";
      break;
    }

  w = Py_BuildValue("(sO)", msg, v);
  Py_DECREF(v);
  PyErr_SetObject(PyExc_SyntaxError, w);
  Py_DECREF(w);

  return NULL;
}


static PyObject *
PrintNonterminals(go, args)
	GrammarObject *go;
	PyObject *args;
{
  char *filename;
  FILE *fp;

  if (! PyArg_ParseTuple(args, "s", &filename))
    return NULL;

  fp = fopen(filename, "w");
  if (fp == NULL)
    {
      PyErr_SetFromErrno(PyExc_IOError);
      return NULL;
    }

  printnonterminals(go->gram.grammar, fp);

  Py_INCREF(Py_None);
  return Py_None;
}


static PyObject *
PrintGrammar(go, args)
     GrammarObject *go;
     PyObject *args;
{
  char *filename;
  FILE *fp;

  if (!PyArg_ParseTuple(args, "s", &filename))
    return NULL;

  fp = fopen(filename, "w");
  if (fp == NULL)
    {
      PyErr_SetFromErrno(PyExc_IOError);
      return NULL;
    }

  printgrammar(go->gram.grammar, fp);

  Py_INCREF(Py_None);
  return Py_None;
}


static PyObject *
ParseFile(go, args)
     GrammarObject *go;
     PyObject *args;
{
  char *filename;
  FILE *fp;
  node *n = NULL;
  PyObject *res;
  perrdetail err;
  int start_state = -1;
  PyObject *so;
  PyObject *start = Py_None;

  if (! PyArg_ParseTuple(args, "s|S", &filename, &start)) 
    return NULL;

  if (start == Py_None)
    start_state = go->gram.grammar->g_start;
  else
    {
      start_state = PySequence_Index(go->gram.nonterminal_list, start);
      if (start_state == -1)
        return 0;
      start_state = start_state + NT_OFFSET;
    }

  /* XXX this needs to allow a fileobject to be passed as well */
  fp = fopen(filename, "r");
  if (fp == NULL)
    {
      PyErr_SetFromErrno(PyExc_IOError);
      return NULL;
    }

  n = PyParser_ParseFile(fp, filename, go->gram.grammar, start_state,
                         (char *)0, (char *)0, &err); 
  fclose(fp);

  if (n == NULL)
    return ParseError(&err);

  so = PyString_FromString(filename);
  res = Node2Tuple(go, so, n);
  PyNode_Free(n);
  Py_XDECREF(so);

  return res;
}


static PyObject *
ParseString(go, args)
     GrammarObject *go;
     PyObject *args;
{
  node *n = NULL;
  PyObject *res;
  perrdetail err;
  int start_state = -1;
  char *string;
  PyObject *start = Py_None;

  if (! PyArg_ParseTuple(args, "s|S", &string, &start))
    return NULL;

  if (start == Py_None)
    start_state = go->gram.grammar->g_start;
  else
    {
      start_state = PySequence_Index(go->gram.nonterminal_list, start);
      if (start_state == -1)
        return NULL;
      start_state = start_state + NT_OFFSET;
    }

  /* XXX I don't know why this is here... perhaps there is a memory
     error somewher in the parser?  */
  string = strdup(string);
  
  n = PyParser_ParseString(string, go->gram.grammar, start_state, &err); 
  if (n == NULL)
    return ParseError(&err);

  res = Node2Tuple(go, "<string>", n);
  PyNode_Free(n);

  return res;
}


static PyObject *
ParseGrammarFile(self, args)
     PyObject *self;
     PyObject *args;
{
  node *n;
  grammar *g0, *g;
  perrdetail err;
  char *filename;
  FILE *fp;
  extern int Py_DebugFlag;

  Py_DebugFlag = 0;

  if (! PyArg_ParseTuple(args, "s|i", &filename, &Py_DebugFlag))
    return NULL;

  fp = fopen(filename, "r");
  if (fp == NULL)
    {
      PyErr_SetFromErrno(PyExc_IOError);
      return NULL;
    }

  g0 = (grammar*) meta_grammar();
  n = PyParser_ParseFile(fp, filename, g0, g0->g_start,
                         (char *)NULL, (char *)NULL, &err);
  fclose(fp);
  /* listtree(n); */

  if (n == NULL)
    return ParseError(&err);

  Py_DebugFlag = 0;
  
  return CreateGrammar(n);
}


static PyObject *
ParseGrammarString(self, args)
     PyObject *self;
     PyObject *args;
{
  node *n;
  grammar *g0, *g;
  perrdetail err;
  char *string;
  extern int Py_DebugFlag;

  Py_DebugFlag = 0;

  if (! PyArg_ParseTuple(args, "s|i", &string, &Py_DebugFlag))
    return NULL;

  g0 = (grammar*) meta_grammar();
  string = strdup(string);
  n = PyParser_ParseString(string, g0, g0->g_start, &err);
  if (n == NULL)
    return ParseError(&err);

  Py_DebugFlag = 0;

  return CreateGrammar(n);
}


/*
 | XXX We don't really want to keep the tree around, but the grammar 
 | generated by pgen references the strings that it contains.  So 
 | until this is fixed...
 */
static PyObject *
CreateGrammar(n)
     node *n;
{
  grammar *g;
  GrammarObject *go;
  PyObject *keyword_list;

  g = (grammar*) pgen(n);
  if (g == NULL)
    {
      PyErr_SetString(PyExc_TypeError, "invalid grammar");
      return NULL;
    }

  go = PyObject_NEW(GrammarObject, &Grammar_Type);
  if (! go)
    return NULL;

  go->gram.grammar = g;
  go->gram.tree = n;
  go->gram.nonterminal_list = CreateNonterminalList(g);
  go->gram.nonterminals = CreateEnum("NT", go->gram.nonterminal_list);
  keyword_list = CreateKeywordList(g);
  go->gram.keywords = CreateEnum("KW", keyword_list);
  Py_DECREF(keyword_list);
  go->gram.reductions = PyDict_New();
  
  return (PyObject *) go;
}


static PyObject *
GetAttr(go, name)
     GrammarObject *go;
     char *name;
{
  if (strcmp(name, "__members__") == 0)
    return Py_BuildValue("(ssss)", "KW", "NT", "T", "reductions");
  
  if (strcmp(name, "reduction") == 0)
    {
      Py_INCREF(go->gram.reductions);
      return go->gram.reductions;
    }
  
  if (strcmp(name, "NT") == 0)
    {
      Py_INCREF(go->gram.nonterminals);
      return go->gram.nonterminals;
    }
  
  if (strcmp(name, "KW") == 0)
    {
      Py_INCREF(go->gram.keywords);
      return go->gram.keywords;
    }
  
  if (strcmp(name, "T") == 0)
    {
      Py_INCREF(TerminalEnum);
      return TerminalEnum;
    }
  
  return Py_FindMethod(GrammarMethods, (PyObject*)go, name);
}


static void
Dealloc (go)
     GrammarObject *go;
{
  Py_DECREF(go->gram.nonterminals);
  Py_DECREF(go->gram.keywords);
  Py_DECREF(go->gram.nonterminal_list);
  Py_DECREF(go->gram.reductions);
                        
  PyNode_Free(go->gram.tree);

  /* XXX really should free the grammar itself */

  free(go);
}


void
initpgen()
{
  PyObject *m;
  PyObject *d;
  PyObject *l;

  m = Py_InitModule("pgen", GlobalMethods);
  d = PyModule_GetDict(m);
  
  TerminalList = CreateTerminalList(_PyParser_TokenNames, N_TOKENS); 
  TerminalEnum = CreateEnum("TERMINALS", TerminalList);
}
